package sun.rmi.server;

import java.io.ObjectStreamClass;
import java.util.HashMap;
import java.util.Collections;
import java.util.WeakHashMap;
import sun.security.action.GetBooleanAction;
import java.rmi.server.RemoteRef;
import java.io.PrintStream;
import java.util.Date;
import java.rmi.server.Operation;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.rmi.ServerError;
import java.rmi.ServerException;
import java.rmi.MarshalException;
import java.lang.reflect.InvocationTargetException;
import java.io.IOException;
import java.rmi.AccessException;
import sun.rmi.transport.StreamRemoteCall;
import java.rmi.UnmarshalException;
import java.rmi.server.RemoteCall;
import java.rmi.server.SkeletonNotFoundException;
import java.rmi.server.ServerNotActiveException;
import sun.rmi.transport.tcp.TCPTransport;
import sun.rmi.transport.Target;
import java.rmi.server.ExportException;
import java.rmi.RemoteException;
import java.rmi.server.RemoteStub;
import java.rmi.Remote;
import sun.rmi.transport.LiveRef;
import java.util.concurrent.atomic.AtomicInteger;
import java.lang.reflect.Method;
import java.util.Map;
import sun.misc.ObjectInputFilter;
import java.rmi.server.Skeleton;
import sun.rmi.runtime.Log;
import java.rmi.server.ServerRef;

public class UnicastServerRef extends UnicastRef implements ServerRef, Dispatcher
{
    public static final boolean logCalls;
    public static final Log callLog;
    private static final long serialVersionUID = -7384275867073752268L;
    private static final boolean wantExceptionLog;
    private boolean forceStubUse;
    private static final boolean suppressStackTraces;
    private transient Skeleton skel;
    private final transient ObjectInputFilter filter;
    private transient Map<Long, Method> hashToMethod_Map;
    private static final WeakClassHashMap<Map<Long, Method>> hashToMethod_Maps;
    private static final Map<Class<?>, ?> withoutSkeletons;
    private final AtomicInteger methodCallIDCount;
    
    public UnicastServerRef() {
        this.forceStubUse = false;
        this.hashToMethod_Map = null;
        this.methodCallIDCount = new AtomicInteger(0);
        this.filter = null;
    }
    
    public UnicastServerRef(LiveRef liveRef) {
        super(liveRef);
        this.forceStubUse = false;
        this.hashToMethod_Map = null;
        this.methodCallIDCount = new AtomicInteger(0);
        this.filter = null;
    }
    
    public UnicastServerRef(LiveRef liveRef, ObjectInputFilter filter) {
        super(liveRef);
        this.forceStubUse = false;
        this.hashToMethod_Map = null;
        this.methodCallIDCount = new AtomicInteger(0);
        this.filter = filter;
    }
    
    public UnicastServerRef(int port) {
        super(new LiveRef(port));
        this.forceStubUse = false;
        this.hashToMethod_Map = null;
        this.methodCallIDCount = new AtomicInteger(0);
        this.filter = null;
    }
    
    public UnicastServerRef(boolean forceStubUse) {
        this(0);
        this.forceStubUse = forceStubUse;
    }
    
    @Override
    public RemoteStub exportObject(Remote remote, Object o) throws RemoteException {
        this.forceStubUse = true;
        return (RemoteStub)this.exportObject(remote, o, false);
    }
    
    public Remote exportObject(Remote remote, Object o, boolean permanent) throws RemoteException {
        Class<Object> class1 = remote.getClass();
        Remote proxy;
        try {
            proxy = Util.createProxy(class1, this.getClientRef(), this.forceStubUse);
        }
        catch (IllegalArgumentException ex) {
            throw new ExportException("remote object implements illegal remote interface", ex);
        }
        if ((proxy instanceof RemoteStub)) {
            this.setSkeleton(remote);
        }
        this.ref.exportObject(new Target(remote, this, proxy, this.ref.getObjID(), permanent));
        this.hashToMethod_Map = (Map)UnicastServerRef.hashToMethod_Maps.get(class1);
        return proxy;
    }
    
    @Override
    public String getClientHost() throws ServerNotActiveException {
        return TCPTransport.getClientHost();
    }
    
    public void setSkeleton(Remote remote) throws RemoteException {
        if (!UnicastServerRef.withoutSkeletons.containsKey(remote.getClass())) {
            try {
                this.skel = Util.createSkeleton(remote);
            }
            catch (SkeletonNotFoundException ex) {
                UnicastServerRef.withoutSkeletons.put(remote.getClass(), null);
            }
        }
    }
    
    @Override
    public void dispatch(Remote obj, RemoteCall remoteCall) throws IOException {
        try {
            ObjectInput inputStream;
            int int1;
            try {
                inputStream = remoteCall.getInputStream();
                int1 = inputStream.readInt();
            }
            catch (Exception ex) {
                throw new UnmarshalException("error unmarshalling call header", ex);
            }
            if (this.skel != null) {
                this.oldDispatch(obj, remoteCall, int1);
                return;
            }
            if (int1 >= 0) {
                throw new UnmarshalException("skeleton class not found but required for client version");
            }
            long long1;
            try {
                long1 = inputStream.readLong();
            }
            catch (Exception ex2) {
                throw new UnmarshalException("error unmarshalling call header", ex2);
            }
            MarshalInputStream marshalInputStream = (MarshalInputStream)inputStream;
            marshalInputStream.skipDefaultResolveClass();
            Method method = (Method)this.hashToMethod_Map.get(Long.valueOf(long1));
            if (method == null) {
                throw new UnmarshalException("unrecognized method hash: method not supported by remote object");
            }
            this.logCall(obj, method);
            Object[] unmarshalParameters = null;
            try {
                this.unmarshalCustomCallData(inputStream);
                unmarshalParameters = this.unmarshalParameters(obj, method, marshalInputStream);
            }
            catch (AccessException ex3) {
                ((StreamRemoteCall)remoteCall).discardPendingRefs();
                throw ex3;
            }
            catch (IOException | ClassNotFoundException ex4) {
                ((StreamRemoteCall)remoteCall).discardPendingRefs();
                throw new UnmarshalException("error unmarshalling arguments", ex4);
            }
            finally {
                remoteCall.releaseInputStream();
            }
            Object invoke;
            try {
                invoke = method.invoke(obj, unmarshalParameters);
            }
            catch (InvocationTargetException ex5) {
                throw ex5.getTargetException();
            }
            try {
                ObjectOutput resultStream = remoteCall.getResultStream(true);
                Class<Object> returnType = method.getReturnType();
                if (returnType != Void.TYPE) {
                    UnicastRef.marshalValue(returnType, invoke, resultStream);
                }
            }
            catch (IOException ex6) {
                throw new MarshalException("error marshalling return", ex6);
            }
        }
        catch (Throwable cause) {
            this.logCallException(cause);
            ObjectOutput resultStream2 = remoteCall.getResultStream(false);
            if ((cause instanceof Error)) {
                cause = new ServerError("Error occurred in server thread", (Error)cause);
            }
            else if ((cause instanceof RemoteException)) {
                cause = new ServerException("RemoteException occurred in server thread", (Exception)cause);
            }
            if (UnicastServerRef.suppressStackTraces) {
                clearStackTraces(cause);
            }
            resultStream2.writeObject(cause);
            if ((cause instanceof AccessException)) {
                throw new IOException("Connection is not reusable", cause);
            }
            remoteCall.releaseInputStream();
            remoteCall.releaseOutputStream();
        }
        finally {
            remoteCall.releaseInputStream();
            remoteCall.releaseOutputStream();
        }
    }
    
    protected void unmarshalCustomCallData(ObjectInput objectInput) throws IOException, ClassNotFoundException {
        if (this.filter != null && objectInput instanceof ObjectInputStream) {
            AccessController.doPrivileged(new PrivilegedAction<Void>() {
                
                @Override
                public Void run() {
                    ObjectInputFilter.Config.setObjectInputFilter(ois, UnicastServerRef.this.filter);
                    return null;
                }
            });
        }
    }
    
    private void oldDispatch(Remote remote, RemoteCall remoteCall, int i) throws Exception {
        ObjectInput inputStream = remoteCall.getInputStream();
        try {
            if (Class.forName("sun.rmi.transport.DGCImpl_Skel").isAssignableFrom(this.skel.getClass())) {
                ((MarshalInputStream)inputStream).useCodebaseOnly();
            }
        }
        catch (ClassNotFoundException ex) {}
        long long1;
        try {
            long1 = inputStream.readLong();
        }
        catch (Exception ex2) {
            throw new UnmarshalException("error unmarshalling call header", ex2);
        }
        Operation[] operations = this.skel.getOperations();
        this.logCall(remote, (i >= 0 && i < operations.length) ? operations[i] : ("op: " + i));
        this.unmarshalCustomCallData(inputStream);
        this.skel.dispatch(remote, remoteCall, i, long1);
    }
    
    public static void clearStackTraces(Throwable cause) {
        StackTraceElement[] stackTrace = new StackTraceElement[0];
        while (cause != null) {
            cause.setStackTrace(stackTrace);
            cause = cause.getCause();
        }
    }
    
    private void logCall(Remote remote, Object obj) {
        if (UnicastServerRef.callLog.isLoggable(Log.VERBOSE)) {
            String clientHost;
            try {
                clientHost = this.getClientHost();
            }
            catch (ServerNotActiveException ex) {
                clientHost = "(local)";
            }
            UnicastServerRef.callLog.log(Log.VERBOSE, "[" + clientHost + ": " + remote.getClass().getName() + this.ref.getObjID().toString() + ": " + obj + "]");
        }
    }
    
    private void logCallException(Throwable t) {
        if (UnicastServerRef.callLog.isLoggable(Log.BRIEF)) {
            String string = "";
            try {
                string = "[" + this.getClientHost() + "] ";
            }
            catch (ServerNotActiveException ex) {}
            UnicastServerRef.callLog.log(Log.BRIEF, string + "exception: ", t);
        }
        if (UnicastServerRef.wantExceptionLog) {
            PrintStream err = System.err;
            synchronized (err) {
                err.println();
                err.println("Exception dispatching call to " + this.ref.getObjID() + " in thread \"" + Thread.currentThread().getName() + "\" at " + new Date() + ":");
                t.printStackTrace(err);
            }
        }
    }
    
    @Override
    public String getRefClass(ObjectOutput objectOutput) {
        return "UnicastServerRef";
    }
    
    protected RemoteRef getClientRef() {
        return new UnicastRef(this.ref);
    }
    
    @Override
    public void writeExternal(ObjectOutput objectOutput) throws IOException {
    }
    
    @Override
    public void readExternal(ObjectInput objectInput) throws IOException, ClassNotFoundException {
        this.ref = null;
        this.skel = null;
    }
    
    private Object[] unmarshalParameters(Object o, Method method, MarshalInputStream marshalInputStream) throws IOException, ClassNotFoundException {
        return (o instanceof DeserializationChecker) ? this.unmarshalParametersChecked((DeserializationChecker)o, method, marshalInputStream) : this.unmarshalParametersUnchecked(method, marshalInputStream);
    }
    
    private Object[] unmarshalParametersUnchecked(Method method, ObjectInput objectInput) throws IOException, ClassNotFoundException {
        Class[] parameterTypes = method.getParameterTypes();
        Object[] array = new Object[parameterTypes.length];
        for (int i = 0; i < parameterTypes.length; ++i) {
            array[i] = UnicastRef.unmarshalValue(parameterTypes[i], objectInput);
        }
        return array;
    }
    
    private Object[] unmarshalParametersChecked(DeserializationChecker deserializationChecker, Method method, MarshalInputStream marshalInputStream) throws IOException, ClassNotFoundException {
        int andIncrement = this.methodCallIDCount.getAndIncrement();
        UnicastServerRef.MyChecker streamChecker = new UnicastServerRef.MyChecker(deserializationChecker, method, andIncrement);
        marshalInputStream.setStreamChecker(streamChecker);
        try {
            Class[] parameterTypes = method.getParameterTypes();
            Object[] array = new Object[parameterTypes.length];
            for (int i = 0; i < parameterTypes.length; array[i] = UnicastRef.unmarshalValue(parameterTypes[i], marshalInputStream), ++i) {
                streamChecker.setIndex(i);
            }
            streamChecker.end(andIncrement);
            return array;
        }
        finally {
            marshalInputStream.setStreamChecker(null);
        }
    }
    
    static {
        logCalls = (boolean)AccessController.doPrivileged(new GetBooleanAction("java.rmi.server.logCalls"));
        callLog = Log.getLog("sun.rmi.server.call", "RMI", UnicastServerRef.logCalls);
        wantExceptionLog = (boolean)AccessController.doPrivileged(new GetBooleanAction("sun.rmi.server.exceptionTrace"));
        suppressStackTraces = (boolean)AccessController.doPrivileged(new GetBooleanAction("sun.rmi.server.suppressStackTraces"));
        hashToMethod_Maps = new UnicastServerRef.HashToMethod_Maps();
        withoutSkeletons = Collections.synchronizedMap(new WeakHashMap());
    }
    
    private static class HashToMethod_Maps extends WeakClassHashMap<Map<Long, Method>>
    {
        HashToMethod_Maps() {
        }
        
        @Override
        protected Map<Long, Method> computeValue(Class<?> superclass) {
            HashMap<Long, Method> hashMap = new HashMap();
            Class[] interfaces;
            int length;
            int i;
            Class<Object> clazz;
            Method[] methods;
            int length2;
            int j;
            Method method;
            for (; superclass != null; superclass = superclass.getSuperclass()) {
                interfaces = superclass.getInterfaces();
                for (length = interfaces.length, i = 0; i < length; ++i) {
                    clazz = interfaces[i];
                    if (Remote.class.isAssignableFrom(clazz)) {
                        methods = clazz.getMethods();
                        for (length2 = methods.length, j = 0; j < length2; ++j) {
                            method = (method = methods[j]);
                            AccessController.doPrivileged(new PrivilegedAction<Void>() {
                                @Override
                                public Void run() {
                                    method.setAccessible(true);
                                    return null;
                                }
                            });
                            hashMap.put(Long.valueOf(Util.computeMethodHash(method)), method);
                        }
                    }
                }
            }
            return hashMap;
        }
    }
    
    private static class MyChecker implements MarshalInputStream.StreamChecker
    {
        private final DeserializationChecker descriptorCheck;
        private final Method method;
        private final int callID;
        private int parameterIndex;
        
        MyChecker(DeserializationChecker descriptorCheck, Method method, int callID) {
            this.descriptorCheck = descriptorCheck;
            this.method = method;
            this.callID = callID;
        }
        
        public void validateDescriptor(ObjectStreamClass objectStreamClass) {
            this.descriptorCheck.check(this.method, objectStreamClass, this.parameterIndex, this.callID);
        }
        
        @Override
        public void checkProxyInterfaceNames(String[] array) {
            this.descriptorCheck.checkProxyClass(this.method, array, this.parameterIndex, this.callID);
        }
        
        void setIndex(int parameterIndex) {
            this.parameterIndex = parameterIndex;
        }
        
        void end(int n) {
            this.descriptorCheck.end(n);
        }
    }
}
